from functools import partial
import torch
import torch.nn.functional as F
from baseline.graphnas import GraphNASAlgorithm
from ogb.nodeproppred import PygNodePropPredDataset
from .supernet import ARXIV_SIMPLE
from supernet.supernet import Supernet
import os
import torch_geometric.transforms as T
from tqdm import tqdm

def gen_model(arch, inputs, outputs, args):
    return Supernet(inputs, args.n_hidden, outputs, args.num_layers, args.dropout, track=True, arch=arch).cuda()

def arch2score(arch, data, inc, ouc, args):
    model = gen_model(arch, inc, ouc, args)
    # train 100 epoch, find the best
    best_acc = 0.
    optimizer = torch.optim.Adam(model.parameters(), lr=args.lr)
    for e in tqdm(range(args.epochs)):
        model.train()
        optimizer.zero_grad()
        out = model(data.x, data.adj_t, arch)[data.train_mask]
        loss = F.nll_loss(out, data.y[data.train_mask])
        loss.backward()
        optimizer.step()

        # evaluate
        model.eval()
        out = model(data.x, data.adj_t, arch)
        val_acc = (out[data.valid_mask].argmax(dim=1) == data.y[data.valid_mask]).float().mean()
        if val_acc > best_acc:
            best_acc = val_acc
            # test_acc = (out[data.test_mask].argmax(dim=1) == data.y[data.test_mask]).float().mean()

    return best_acc

if __name__ == '__main__':
    import argparse
    import time
    begin_time = time.time()
    parser = argparse.ArgumentParser(description='graphnas-coauthor')
    parser.add_argument('--device', type=str, default="cuda")
    parser.add_argument('--num-layers', type=int, default=3)
    parser.add_argument('--n-hidden', type=int, default=256)

    parser.add_argument('--dropout', type=float, default=0.5)
    parser.add_argument('--lr', type=float, default=0.01)
    parser.add_argument('--epochs', type=int, default=500)
    parser.add_argument('--gnas-lr', type=float, default=0.01)
    parser.add_argument('--T', type=float, default=1)

    args = parser.parse_args()
    print(args)

    device = args.device

    dataset = PygNodePropPredDataset(name='ogbn-arxiv', root=os.path.expanduser("~/dataset/pyg"), transform=T.ToSparseTensor())
    
    data = dataset[0]
    data.adj_t = data.adj_t.to_symmetric()
    data = data.to(device)
    data.y = data.y[:,0]

    split_idx = dataset.get_idx_split()
    train_idx = split_idx['train'].to(device)
    valid_idx = split_idx['valid'].to(device)
    test_idx = split_idx['test'].to(device)

    data.train_mask = train_idx
    data.valid_mask = valid_idx
    data.test_mask = test_idx

    graphnas = GraphNASAlgorithm(ARXIV_SIMPLE, args.num_layers, args.T, args.gnas_lr, args.device)
    graphnas.train_constraint(partial(arch2score, data=data, args=args, inc=data.x.size(1), ouc=dataset.num_classes), archs=20)
    graphnas.evaluate()
